/*
 * main.cpp
 * Copyright (C) Aitor Cuadrado Zubizarreta <aitor_czr@gnuinos.org>
 * 
 * simple-netaid is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * simple-netaid is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * See the COPYING file. * 
 */
   

#include <iostream>
#include <memory>
#include <vector>
#include <stdbool.h>
#include <signal.h>
#include <unistd.h>
#include <cassert>
#include <time.h>

#include <globals.h>
#include "def.h"
#include "cdk_main.h"
#include "functions_ui.h"

#define MAX_SIZE 1024
#define WHITELIST "/etc/simple-netaid/whitelist"

#define strdup_or_null(str)  (str) != NULL ? strdup(str) : NULL

functions_Ui * f_ui;

C_LINKAGE_BEGIN
#include <sigact.h>
#include <simple-netaid/sbuf.h>
#include <simple-netaid/helpers.h>
#include <simple-netaid/iproute.h>
   
void emit_sigusr(int, const char*);
void send_arguments(int, int, ...);
void wait_on_run();
void get_interfaces(struct sbuf*, struct sbuf*);
void interface_up(const char*);
void interface_down(const char*);
void ipaddr_flush(const char*);
void get_pstat_binary(struct sbuf*, const char*);
char* strncpy_t(char*, size_t, const char*, size_t);

functions_Ui* new_functions_ui() {
   return new functions_Ui();
}
  
void ui_cleanup() {
   return f_ui->do_cleanup();
}
  
void ui_display() {
   return f_ui->do_display();
}
  
void ui_delete() {
   return f_ui->do_delete();
}
C_LINKAGE_END

char arg0[256] = {0};
const char *libpath;
const char *progname="simple-netaid-cdk";
int p = 0;

char m_essid[256] = {0};
char m_passwd[256] = {0};
char m_filename[256] = {0};

using namespace std;

void loopMain(int*, string, string);
void query_installed_binary(const char*, struct sbuf*);
bool query_whitelist(const char*);
void is_already_running(const char*);

// Get absolute path to libnetaid.so.*
static void get_fullpath(struct sbuf *s, const char *name)
{
   char *token;
   char delim[4] = "=> ";
   char line[1024] = {0};
   char cmd[1024] = {0};
   
   sprintf(cmd, "ldd ");
   strcat(cmd, name);
   strcat(cmd, " | grep libnetaid.so");
   
   FILE *pf = popen(cmd, "r");
   if(!pf) {
      printf("Sorry, there was an error opening the pipe\n");
      exit(EXIT_FAILURE);
   }
   
   fgets(line, 1024, pf);
   token = strtok((char*)line, delim);
   token = strtok(NULL, delim);
   sbuf_addstr(s, token);
   pclose(pf);
}

// TODO: define the arguments in the command line
// atoi(argv[1]) == 0 -> monitor
int main(int argc, char *argv[])
{             
   struct sbuf s, path;   
   struct sbuf wired, wireless;
   char *arr = NULL;
   pid_t pid;
   
   signal(SIGUSR2, SIG_IGN);
   
   strncpy_t(arg0, 256, argv[0], strlen(argv[0]) + 1);    
   if(argc == 1) argv[1] = strdup("0");      
   
   // Get our pid:
   pid = getpid();
   if (asprintf(&arr, "%jd", (intmax_t) pid) == -1) {
      printf("Unable to get the pid.\nPress Any Key to Exit...\n");
      getchar();
      exit(EXIT_FAILURE);
   }
     
   is_already_running(arr);
   
   sbuf_init(&path);
   query_installed_binary(arg0, &path);   
   if(path.len == 0) {
      bool whitelisted; 
      get_pstat_binary(&path, arr);
      free(arr);
   }
      
   if(path.len > 0) {
      bool whitelisted = query_whitelist(path.buf);
      if(!whitelisted) {       
         printf("\n'%s' is not included in '/etc/simple-netaid/whitelist'.\nPress Any Key to Exit...\n", path.buf);
         getchar();
         exit(EXIT_FAILURE);
      }
   }
   else {
	  printf("Unable to get the path of %s\n.Press Any Key to Exit...\n", arg0);
      getchar();
      exit(EXIT_FAILURE);
   }
     
   printf("Running %s...\n", path.buf);
    
   // Get the fullpath to libnetaid
   sbuf_init(&s);
   get_fullpath(&s, path.buf);
   libpath = strdup_or_null(s.buf);
   sbuf_free(&s);
   sbuf_free(&path);
   
   // internationalization
   {
       int j;
       const char *var[3] = {"LC_MESSAGES", "LC_ALL", "LANG"};
       const char *value, *mylocale;
       for(j=0; j<3; j++)
         {
            value=getenv(var[j]);
            /* Unset */
            setenv(var[j], "", 1);
            if(value) break;
         }
       setenv("LC_CTYPE", "", 1);
       mylocale = setlocale(LC_MESSAGES, value);    
       bindtextdomain(progname, "/usr/share/locale");
       textdomain(progname);
   }
 
   // Does nothing, just in order to hold the daemon our pid
   if(argc == 1) {
      emit_sigusr((-1)*pid, libpath);
   }
	
   sbuf_init(&wired);
   sbuf_init(&wireless);
   get_interfaces(&wired, &wireless);
   
   string wired_device = string(wired.buf);
   string wireless_device = string(wireless.buf);
   
   sbuf_free(&wired);
   sbuf_free(&wireless);
   
   f_ui = new functions_Ui();   
   CdkMain *cdk_main = new CdkMain(wired_device, wireless_device); 
 
   for ( ;; ) {
      loopMain(&p, wired_device, wireless_device);
      if(p == QUIT) break;
   }
   
   delete f_ui;
   delete cdk_main;
   emit_sigusr(-1, libpath);
   
   endCDK();
   ExitProgram(EXIT_SUCCESS);
}

void loopMain(int *p, string wired_device, string wireless_device)
{ 
   switch (*p) {
	  case QUIT:
	     return;  /* end the loop */
      case WIRED_UP:    
         emit_sigusr(getpid(), libpath);
         send_arguments(2, 1, wired_device.c_str()); 
         wait_for_sigusr1();
         break;
      case WIRED_DOWN:    
         emit_sigusr(getpid(), libpath);
         send_arguments(3, 1, wired_device.c_str()); 
         wait_for_sigusr1();
         break;
      case WIRED_DISCONNECT:
         emit_sigusr(getpid(), libpath);
         send_arguments(0, 1, wired_device.c_str());
         wait_for_sigusr1();
         break;
      case WIRED_CONNECTION:     
         emit_sigusr(getpid(), libpath);
         send_arguments(1, 1, wired_device.c_str());
         wait_for_sigusr1();
         {
            int count = 0;
            while(*iproute() == 0 && count < 60) {
			   sleep(1);
			   count++;
			}
         }
         break;
      case WIRELESS_UP:    
         emit_sigusr(getpid(), libpath);
         send_arguments(2, 1, wireless_device.c_str());
         wait_for_sigusr1();
         break;
      case WIRELESS_DOWN:   
         emit_sigusr(getpid(), libpath);
         send_arguments(3, 1, wireless_device.c_str()); 
         wait_for_sigusr1();
         break;
      case WIRELESS_DISCONNECT:
         emit_sigusr(getpid(), libpath);
         send_arguments(0, 1, wireless_device.c_str());
         wait_for_sigusr1();
         break;
      case WIRELESS_CONNECTION:    
         emit_sigusr(getpid(), libpath);
         send_arguments(4, 3, wireless_device.c_str(), m_essid, m_passwd);
         wait_for_sigusr1();
         {
            int count = 0;
            while(*iproute() == 0 && count < 60) {
			   sleep(1);
			   count++;
			}
         }
         break;
      case INSTALL_AND_CONNECT:
         {    
            emit_sigusr(getpid(), libpath);
            string filename = "/etc/network/wifi/" + string(m_filename);
            send_arguments(5, 4, wireless_device.c_str(), m_essid, m_passwd, filename.c_str());
            wait_for_sigusr1();
            int count = 0;
            while(*iproute() == 0 && count < 60) {
			   sleep(1);
			   count++;
			}
         }
         break; 
      case CONNECT_TO_SAVED:    
         {
		    emit_sigusr(getpid(), libpath);
            send_arguments(6, 2, wireless_device.c_str(), m_filename);
            wait_for_sigusr1();
         }
         break;
         
      case UNINSTALL: 
         emit_sigusr(getpid(), libpath);
         send_arguments(7, 1, m_filename);
         break;               
     default:
         break;
   }
   
   sigaction_init();

   if(*p > QUIT && *p < UNINSTALL) {
      fd_set fds;
      time_t t = 0;		 
      struct timeval tv;
   
      sleep(1);
   
Select:
      FD_ZERO(&fds);    
       	
      tv.tv_sec = 2;
      tv.tv_usec = 0;
		
      if (t) {
         int delay = t - time(NULL);
         if (delay < 0)
            tv.tv_sec = 0;
         else if (delay < tv.tv_sec)
            tv.tv_sec = delay;
      } 

      if (select(FD_SETSIZE, &fds, NULL, NULL, &tv) < 0) {
         if (errno == EINTR) 
            goto Select;
      }
   }

   ui_display();
}

void query_installed_binary(const char *cstr, struct sbuf *s)
{
   FILE *pf;
   char line[MAX_SIZE] = {0};
   char cmd[128] = {0};
   
   sprintf(cmd, "command -v ");
   strcat(cmd, cstr);
   
   pf = popen(cmd, "r");
   if(!pf) {
      printf("Pipe error: %s\n", cmd);
      exit(EXIT_FAILURE);
   }
     
   fgets(line, MAX_SIZE, pf);
   line[strcspn(line, "\n")] = 0;
   
   if(query_whitelist(line)) { // We'll get false for relative paths
	  sbuf_addstr(s, line);
   }
   pclose(pf);
}

bool query_whitelist(const char *binary_path)
{
   FILE *wl;
   char line[MAX_SIZE] = {0};
   bool res = false;
   
   wl = fopen(WHITELIST, "r");
   if(!wl) {
      printf("Pipe error: %s\n", WHITELIST);
      exit(EXIT_FAILURE);
   }
     
   while(fgets(line, MAX_SIZE, wl)) {
      line[strcspn(line, "\n")] = 0;
	  char str[MAX_SIZE] = {0};
      sprintf(str, "\"");
      strcat(str, line);
      strcat(str, "\"");        
      // Do they match?
      if(!strcmp(line, binary_path)) {
         res = true;
         break;
      }
   }
   
   fclose(wl); 
   return res;
}

void is_already_running(const char *arr)
{
   FILE *pf;
   char line[64] = {0};
   
   pf = popen("pidof simple-netaid-cdk", "r");
   if(!pf) {
      printf("Pipe error\n");
      exit(EXIT_FAILURE);
   }
   
   int count = 0;  
   fgets(line, 64, pf);
   line[strcspn(line, "\n")] = 0;
   if(strcmp(line, arr)) {
      pclose(pf); 
	  printf("\nMultiple instances of %s running at the same time with pids: %s.\nPress Any Key to Exit...\n", progname, line);
      getchar();
      exit(EXIT_FAILURE);
   }
   
   pclose(pf);
}

